// Copyright (c) Facebook, Inc. and its affiliates.
//
// This source code is licensed under the MIT license found in the
// LICENSE file in the root directory of this source tree.
//
use crate::abstract_value::AbstractValue;
use crate::abstract_value::AbstractValueTrait;
use crate::environment::Environment;
use crate::expression::{Expression, ExpressionType};
use crate::k_limits;

use log_derive::logfn_inputs;
use mirai_annotations::assume;
use rustc::hir::def_id::DefId;
use serde::{Deserialize, Serialize};
use std::collections::hash_map::DefaultHasher;
use std::collections::HashSet;
use std::fmt::{Debug, Formatter, Result};
use std::hash::{Hash, Hasher};
use std::rc::Rc;

/// During join and widen operations, paths are copied from one environment to another, causing them
/// to get rehashed. This turns out to be expensive, so for this case we cache the hash to avoid
/// recomputing it. The caching has a cost, so only use this in cases where it is highly likely
/// that the path will be hashed more than once.
#[derive(Serialize, Deserialize, Clone, Eq, Ord, PartialOrd)]
pub struct Path {
    pub value: PathEnum,
    hash: u64,
}

impl Debug for Path {
    fn fmt(&self, f: &mut Formatter<'_>) -> Result {
        self.value.fmt(f)
    }
}

impl Hash for Path {
    fn hash<H: Hasher>(&self, state: &mut H) {
        state.write_u64(self.hash);
    }
}

impl PartialEq for Path {
    fn eq(&self, other: &Path) -> bool {
        self.hash == other.hash && self.value == other.value
    }
}

impl From<PathEnum> for Path {
    #[logfn_inputs(TRACE)]
    fn from(value: PathEnum) -> Self {
        let mut hasher = DefaultHasher::new();
        value.hash(&mut hasher);
        Path {
            value,
            hash: hasher.finish(),
        }
    }
}

/// A path represents a left hand side expression.
/// When the actual expression is evaluated at runtime it will resolve to a particular memory
/// location. During analysis it is used to keep track of state changes.
#[derive(Serialize, Deserialize, Clone, Eq, PartialEq, Hash, Ord, PartialOrd)]
pub enum PathEnum {
    /// A dynamically allocated memory block.
    AbstractHeapAddress { ordinal: usize },

    /// Sometimes a constant value needs to be treated as a path during refinement.
    /// Don't use this unless you are really sure you know what you are doing.
    Constant { value: Rc<AbstractValue> },

    /// 0 is the return value temporary
    /// [1 ... arg_count] are the parameters
    /// [arg_count ... ] are the local variables and compiler temporaries.
    LocalVariable { ordinal: usize },

    /// The name is a summary cache key string.
    StaticVariable {
        /// The crate specific key that is used to identify the function in the current crate.
        /// This is not available for functions returned by calls to functions from other crates,
        /// since the def id the other crates use have no meaning for the current crate.
        #[serde(skip)]
        def_id: Option<DefId>,
        /// The key to use when retrieving a summary for the static variable from the summary cache.
        summary_cache_key: Rc<String>,
        /// The type to use when the static variable value is not yet available.
        expression_type: ExpressionType,
    },

    /// The ordinal is an index into a method level table of MIR bodies.
    /// This should not be serialized into a summary since it is function private local state.
    PromotedConstant { ordinal: usize },

    /// The qualifier denotes some reference, struct, or collection.
    /// The selector denotes a de-referenced item, field, or element, or slice.
    QualifiedPath {
        length: usize,
        qualifier: Rc<Path>,
        selector: Rc<PathSelector>,
    },
}

impl Debug for PathEnum {
    fn fmt(&self, f: &mut Formatter<'_>) -> Result {
        match self {
            PathEnum::AbstractHeapAddress { ordinal } => {
                f.write_fmt(format_args!("heap_{}", ordinal))
            }
            PathEnum::Constant { value } => value.fmt(f),
            PathEnum::LocalVariable { ordinal } => f.write_fmt(format_args!("local_{}", ordinal)),
            PathEnum::StaticVariable {
                summary_cache_key, ..
            } => summary_cache_key.fmt(f),
            PathEnum::PromotedConstant { ordinal } => {
                f.write_fmt(format_args!("constant_{}", ordinal))
            }
            PathEnum::QualifiedPath {
                qualifier,
                selector,
                ..
            } => f.write_fmt(format_args!("{:?}.{:?}", qualifier, selector)),
        }
    }
}

impl Path {
    /// True if path qualifies root, or another qualified path rooted by root.
    #[logfn_inputs(TRACE)]
    pub fn is_rooted_by(&self, root: &Rc<Path>) -> bool {
        match &self.value {
            PathEnum::QualifiedPath { qualifier, .. } => {
                *qualifier == *root || qualifier.is_rooted_by(root)
            }
            _ => false,
        }
    }

    /// True is path qualifies an abstract heap address, or another qualified path rooted by an
    /// abstract heap address.
    #[logfn_inputs(TRACE)]
    pub fn is_rooted_by_abstract_heap_address(&self) -> bool {
        match &self.value {
            PathEnum::QualifiedPath { qualifier, .. } => match qualifier.value {
                PathEnum::AbstractHeapAddress { .. } => true,
                _ => false,
            },
            _ => false,
        }
    }

    // Returns the length of the path.
    #[logfn_inputs(TRACE)]
    pub fn path_length(&self) -> usize {
        match &self.value {
            PathEnum::QualifiedPath { length, .. } => *length,
            _ => 1,
        }
    }

    /// Creates a path the selects the discriminant of the enum at the given path.
    #[logfn_inputs(TRACE)]
    pub fn new_discriminant(enum_path: Rc<Path>, environment: &Environment) -> Rc<Path> {
        let selector = Rc::new(PathSelector::Discriminant);
        Self::new_qualified(enum_path, selector).refine_paths(environment)
    }

    /// Creates a path the selects the given field of the struct at the given path.
    #[logfn_inputs(TRACE)]
    pub fn new_field(
        qualifier: Rc<Path>,
        field_index: usize,
        environment: &Environment,
    ) -> Rc<Path> {
        let selector = Rc::new(PathSelector::Field(field_index));
        Self::new_qualified(qualifier, selector).refine_paths(environment)
    }

    /// Creates a path the selects the element at the given index value of the array at the given path.
    #[logfn_inputs(TRACE)]
    pub fn new_index(
        collection_path: Rc<Path>,
        index_value: Rc<AbstractValue>,
        environment: &Environment,
    ) -> Rc<Path> {
        let selector = Rc::new(PathSelector::Index(index_value));
        Self::new_qualified(collection_path, selector).refine_paths(environment)
    }

    /// Creates a path to the local variable corresponding to the ordinal.
    #[logfn_inputs(TRACE)]
    pub fn new_local(ordinal: usize) -> Rc<Path> {
        Rc::new(PathEnum::LocalVariable { ordinal }.into())
    }

    /// Creates a path the selects the length of the array at the given path.
    #[logfn_inputs(TRACE)]
    pub fn new_length(array_path: Rc<Path>, environment: &Environment) -> Rc<Path> {
        let selector = Rc::new(PathSelector::ArrayLength);
        Self::new_qualified(array_path, selector).refine_paths(environment)
    }

    /// Creates a path the selects the length of the string at the given path.
    #[logfn_inputs(TRACE)]
    pub fn new_string_length(string_path: Rc<Path>, environment: &Environment) -> Rc<Path> {
        let selector = Rc::new(PathSelector::StringLength);
        Self::new_qualified(string_path, selector).refine_paths(environment)
    }

    /// Creates a path the selects the given model field of the value at the given path.
    #[logfn_inputs(TRACE)]
    pub fn new_model_field(
        qualifier: Rc<Path>,
        field_name: Rc<String>,
        environment: &Environment,
    ) -> Rc<Path> {
        let selector = Rc::new(PathSelector::ModelField(field_name));
        Self::new_qualified(qualifier, selector).refine_paths(environment)
    }

    /// Creates a path the qualifies the given root path with the given selector.
    #[logfn_inputs(TRACE)]
    pub fn new_qualified(qualifier: Rc<Path>, selector: Rc<PathSelector>) -> Rc<Path> {
        let qualifier_length = qualifier.path_length();
        if qualifier_length >= k_limits::MAX_PATH_LENGTH {
            warn!("max path length exceeded {:?}.{:?}", qualifier, selector);
        }
        assume!(qualifier_length < 1_000_000_000); // We'll run out of memory long before this happens
        Rc::new(
            PathEnum::QualifiedPath {
                qualifier,
                selector,
                length: qualifier_length + 1,
            }
            .into(),
        )
    }

    /// Adds any abstract heap addresses found in embedded index values to the given set.
    #[logfn_inputs(TRACE)]
    pub fn record_heap_addresses(&self, result: &mut HashSet<usize>) {
        match &self.value {
            PathEnum::QualifiedPath {
                qualifier,
                selector,
                ..
            } => {
                (**qualifier).record_heap_addresses(result);
                selector.record_heap_addresses(result);
            }
            PathEnum::AbstractHeapAddress { ordinal } => {
                result.insert(*ordinal);
            }
            _ => (),
        }
    }
}

pub trait PathRefinement: Sized {
    /// Refine parameters inside embedded index values with the given arguments.
    fn refine_parameters(
        &self,
        arguments: &[(Rc<Path>, Rc<AbstractValue>)],
        fresh: usize,
    ) -> Rc<Path>;

    /// Refine paths that reference other paths.
    /// I.e. when a reference is passed to a function that then returns
    /// or leaks it back to the caller in the qualifier of a path then
    /// we want to dereference the qualifier in order to normalize the path
    /// and not have more than one path for the same location.
    fn refine_paths(&self, environment: &Environment) -> Rc<Path>;

    /// Returns a copy path with the root replaced by new_root.
    fn replace_root(&self, old_root: &Rc<Path>, new_root: Rc<Path>) -> Rc<Path>;
}

impl PathRefinement for Rc<Path> {
    /// Refine parameters inside embedded index values with the given arguments.
    #[logfn_inputs(TRACE)]
    fn refine_parameters(
        &self,
        arguments: &[(Rc<Path>, Rc<AbstractValue>)],
        fresh: usize,
    ) -> Rc<Path> {
        match &self.value {
            PathEnum::LocalVariable { ordinal } => {
                if 0 < *ordinal && *ordinal <= arguments.len() {
                    arguments[*ordinal - 1].0.clone()
                } else {
                    Path::new_local(ordinal + fresh)
                }
            }
            PathEnum::QualifiedPath {
                qualifier,
                selector,
                ..
            } => {
                let refined_qualifier = qualifier.refine_parameters(arguments, fresh);
                let refined_selector = selector.refine_parameters(arguments, fresh);
                Path::new_qualified(refined_qualifier, refined_selector)
            }
            _ => self.clone(),
        }
    }

    /// Refine paths that reference other paths and canonicalize the refinements.
    /// I.e. when a reference is passed to a function that then returns
    /// or leaks it back to the caller in the qualifier of a path then
    /// we want to dereference the qualifier in order to normalize the path
    /// and not have more than one path for the same location.
    #[logfn_inputs(TRACE)]
    fn refine_paths(&self, environment: &Environment) -> Rc<Path> {
        if let Some(val) = environment.value_at(&self) {
            // If the environment has self as a key, then self is canonical, since we should only
            // use canonical paths as keys. The value at the canonical key, however, could just
            // be a reference to another path, which is something that happens during refinement.
            return match &val.expression {
                Expression::Variable { path, .. } | Expression::Widen { path, .. } => path.clone(),
                Expression::AbstractHeapAddress(ordinal) => {
                    Rc::new(PathEnum::AbstractHeapAddress { ordinal: *ordinal }.into())
                }
                _ => self.clone(), // self is canonical
            };
        };
        // self is a path that is not a key in the environment. This could be because it is not
        // canonical, which can only be the case if self is a qualified path.
        if let PathEnum::QualifiedPath {
            qualifier,
            selector,
            ..
        } = &self.value
        {
            let refined_selector = selector.refine_paths(environment);
            let refined_qualifier = qualifier.refine_paths(environment);
            // The qualifier is now canonical. But in the context of a selector, we
            // might be able to simplify the qualifier by dropping an explicit dereference
            // or an explicit reference.
            if let PathEnum::QualifiedPath {
                qualifier: base_qualifier,
                selector: base_selector,
                ..
            } = &refined_qualifier.value
            {
                if *base_selector.as_ref() == PathSelector::Deref {
                    // no need for an explicit deref in a qualifier
                    return Path::new_qualified(base_qualifier.clone(), selector.clone());
                }
            }
            if let Some(val) = environment.value_at(&refined_qualifier) {
                if let Expression::Reference(path) = &val.expression {
                    match refined_selector.as_ref() {
                        PathSelector::Deref => {
                            // if selector is a deref we can just drop the &* sequence
                            return path.clone();
                        }
                        _ => {
                            // drop the explicit reference
                            return Path::new_qualified(path.clone(), selector.clone());
                        }
                    }
                }
            }
            Path::new_qualified(refined_qualifier, refined_selector)
        } else {
            self.clone() // Non qualified paths are already canonical
        }
    }

    /// Returns a copy path with the root replaced by new_root.
    #[logfn_inputs(TRACE)]
    fn replace_root(&self, old_root: &Rc<Path>, new_root: Rc<Path>) -> Rc<Path> {
        match &self.value {
            PathEnum::QualifiedPath {
                qualifier,
                selector,
                ..
            } => {
                let new_qualifier = if *qualifier == *old_root {
                    new_root
                } else {
                    qualifier.replace_root(old_root, new_root)
                };
                Path::new_qualified(new_qualifier, selector.clone())
            }
            _ => new_root,
        }
    }
}

/// The selector denotes a de-referenced item, field, or element, or slice.
#[derive(Serialize, Deserialize, Clone, Eq, PartialEq, Hash, Ord, PartialOrd)]
pub enum PathSelector {
    /// The length of an array.
    ArrayLength,

    /// The length of a string.
    StringLength,

    /// Given a path that denotes a reference, select the thing the reference points to.
    Deref,

    /// The tag used to indicate which case of an enum is used for a particular enum value.
    Discriminant,

    /// Select the struct field with the given index.
    Field(usize),

    /// Select the collection element with the index specified by the abstract value.
    Index(Rc<AbstractValue>),

    /// These indices are generated by slice patterns. Easiest to explain
    /// by example:
    ///
    /// ```
    /// [X, _, .._, _, _] => { offset: 0, min_length: 4, from_end: false },
    /// [_, X, .._, _, _] => { offset: 1, min_length: 4, from_end: false },
    /// [_, _, .._, X, _] => { offset: 2, min_length: 4, from_end: true },
    /// [_, _, .._, _, X] => { offset: 1, min_length: 4, from_end: true },
    /// ```
    ConstantIndex {
        /// index or -index (in Python terms), depending on from_end
        offset: u32,
        /// thing being indexed must be at least this long
        min_length: u32,
        /// counting backwards from end?
        from_end: bool,
    },

    /// These indices are generated by slice patterns.
    ///
    /// slice[from:-to] in Python terms.
    Subslice { from: u32, to: u32 },

    /// "Downcast" to a variant of an ADT. Currently, MIR only introduces
    /// this for ADTs with more than one variant. The value is the ordinal of the variant.
    Downcast(usize),

    /// Select the struct model field with the given name.
    /// A model field is a specification construct used during MIRAI verification
    /// and does not have a runtime location.
    ModelField(Rc<String>),
}

impl Debug for PathSelector {
    fn fmt(&self, f: &mut Formatter<'_>) -> Result {
        match self {
            PathSelector::ArrayLength | PathSelector::StringLength => f.write_str("len()"),
            PathSelector::Deref => f.write_str("deref"),
            PathSelector::Discriminant => f.write_str("discr"),
            PathSelector::Field(index) => index.fmt(f),
            PathSelector::Index(value) => f.write_fmt(format_args!("[{:?}]", value)),
            PathSelector::ConstantIndex {
                offset,
                min_length,
                from_end,
            } => f.write_fmt(format_args!(
                "[offset: {}, min_length: {}, from_end: {}",
                offset, min_length, from_end
            )),
            PathSelector::Subslice { from, to } => f.write_fmt(format_args!("[{} : {}]", from, to)),
            PathSelector::Downcast(ordinal) => f.write_fmt(format_args!(" as {}", ordinal)),
            PathSelector::ModelField(name) => name.fmt(f),
        }
    }
}

impl PathSelector {
    /// Adds any abstract heap addresses found in embedded index values to the given set.
    #[logfn_inputs(TRACE)]
    pub fn record_heap_addresses(&self, result: &mut HashSet<usize>) {
        if let PathSelector::Index(value) = self {
            value.record_heap_addresses(result);
        }
    }
}

pub trait PathSelectorRefinement: Sized {
    /// Refine parameters inside embedded index values with the given arguments.
    fn refine_parameters(&self, arguments: &[(Rc<Path>, Rc<AbstractValue>)], fresh: usize) -> Self;

    /// Returns a value that is simplified (refined) by replacing values with Variable(path) expressions
    /// with the value at that path (if there is one). If no refinement is possible
    /// the result is simply a clone of this value. This refinement only makes sense
    /// following a call to refine_parameters.
    fn refine_paths(&self, environment: &Environment) -> Self;
}

impl PathSelectorRefinement for Rc<PathSelector> {
    /// Refine parameters inside embedded index values with the given arguments.
    #[logfn_inputs(TRACE)]
    fn refine_parameters(
        &self,
        arguments: &[(Rc<Path>, Rc<AbstractValue>)],
        fresh: usize,
    ) -> Rc<PathSelector> {
        if let PathSelector::Index(value) = self.as_ref() {
            let refined_value = value.refine_parameters(arguments, fresh);
            Rc::new(PathSelector::Index(refined_value))
        } else {
            self.clone()
        }
    }

    /// Returns a value that is simplified (refined) by replacing values with Variable(path) expressions
    /// with the value at that path (if there is one). If no refinement is possible
    /// the result is simply a clone of this value. This refinement only makes sense
    /// following a call to refine_parameters.
    #[logfn_inputs(TRACE)]
    fn refine_paths(&self, environment: &Environment) -> Rc<PathSelector> {
        if let PathSelector::Index(value) = self.as_ref() {
            let refined_value = value.refine_paths(environment);
            Rc::new(PathSelector::Index(refined_value))
        } else {
            self.clone()
        }
    }
}
